@@ -293,26 +293,27 @@ int set_numberformat(int *numberformat, char *number, int number_type)
 // if udh is true, then udh_data contains the optional user data header in hex dump, example: "05 00 03 AF 02 01"
 
 void make_pdu(char* number, char* message, int messagelen, int alphabet, int flash_sms, int report, int with_udh,
-              char* udh_data, char* mode, char* pdu, int validity, int replace_msg, int system_msg, int number_type)
+              char* udh_data, char* mode, char* pdu, int validity, int replace_msg, int system_msg, int number_type, char *smsc)
 {
   int coding;
   int flags;
-  char tmp[50];
+  char tmp[SIZE_TO];
   char tmp2[500];
   int numberformat;
   int numberlength;
   char *p;
   int l;
+  char tmp_smsc[SIZE_SMSC];
 
   if (number[0] == 's')  // Is number starts with s, then send it without number format indicator
   {
     numberformat = NF_UNKNOWN;
-    strcpy(tmp, number +1);
+    snprintf(tmp, sizeof(tmp), "%s", number +1);
   }
   else
   {
     numberformat = NF_INTERNATIONAL;
-    strcpy(tmp, number);
+    snprintf(tmp, sizeof(tmp), "%s", number);
   }
 
   set_numberformat(&numberformat, tmp, number_type);
@@ -324,6 +325,19 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
   // Swap every second character
   swapchars(tmp);
 
+  // 3.1.12:
+  *tmp_smsc = 0;
+  if (DEVICE.smsc_pdu && (smsc[0] || DEVICE.smsc[0]))
+  {
+    p = (smsc[0]) ? smsc : DEVICE.smsc;
+    while (*p == '+')
+      p++;
+    snprintf(tmp_smsc, sizeof(tmp_smsc), "%s", p);
+    if (strlen(tmp_smsc) % 2 && strlen(tmp_smsc) < sizeof(tmp_smsc) - 1)
+      strcat(tmp_smsc, "F");
+    swapchars(tmp_smsc);
+  }
+
   flags=1; // SMS-Sumbit MS to SMSC
   if (with_udh)
     flags+=64; // User Data Header
@@ -390,7 +404,14 @@ void make_pdu(char* number, char* message, int messagelen, int alphabet, int fla
     //sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X", flags, numberlength, numberformat, tmp,
     // (system_msg) ? 0x40 : (replace_msg >= 1 && replace_msg <= 7) ? 0x40 + replace_msg : 0,
     // (system_msg) ? 0xF4 : coding, validity, messagelen);
-    sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X", flags, numberlength, numberformat, tmp, proto, coding, validity, messagelen);
+
+    // 3.1.12:
+    //sprintf(pdu, "00%02X00%02X%02X%s%02X%02X%02X%02X", flags, numberlength, numberformat, tmp, proto, coding, validity, messagelen);
+    if (*tmp_smsc)
+      sprintf(pdu, "%02X%s%s", (int)strlen(tmp_smsc) / 2 + 1, (tmp_smsc[1] == '0')? "81": "91", tmp_smsc);
+    else
+      strcpy(pdu, "00");
+    sprintf(strchr(pdu, 0), "%02X00%02X%02X%s%02X%02X%02X%02X", flags, numberlength, numberformat, tmp, proto, coding, validity, messagelen);
   }
 
   /* concatenate the text to the PDU string */
@@ -943,6 +964,63 @@ int explain_toa(char *dest, char *octet_char, int octet_int)
   return result;
 }
 
+// 3.1.14:
+void explain_status(char *dest, size_t size_dest, int status)
+{
+  char *p = "unknown";
+
+  switch (status)
+  {
+    case 0: p = "Ok,short message received by the SME"; break;
+    case 1: p = "Ok,short message forwarded by the SC to the SME but the SC is unable to confirm delivery"; break;
+    case 2: p = "Ok,short message replaced by the SC"; break;
+
+    // Temporary error, SC still trying to transfer SM
+    case 32: p = "Still trying,congestion"; break;
+    case 33: p = "Still trying,SME busy"; break;
+    case 34: p = "Still trying,no response sendr SME"; break;
+    case 35: p = "Still trying,service rejected"; break;
+    case 36: p = "Still trying,quality of service not available"; break;
+    case 37: p = "Still trying,error in SME"; break;
+    // 38...47: Reserved
+    // 48...63: Values specific to each SC
+
+    // Permanent error, SC is not making any more transfer attempts
+    case 64: p = "Error,remote procedure error"; break;
+    case 65: p = "Error,incompatible destination"; break;
+    case 66: p = "Error,connection rejected by SME"; break;
+    case 67: p = "Error,not obtainable"; break;
+    case 68: p = "Error,quality of service not available"; break;
+    case 69: p = "Error,no interworking available"; break;
+    case 70: p = "Error,SM validity period expired"; break;
+    case 71: p = "Error,SM deleted by originating SME"; break;
+    case 72: p = "Error,SM deleted by SC administration"; break;
+    case 73: p = "Error,SM does not exist"; break;
+    // 74...79: Reserved
+    // 80...95: Values specific to each SC
+
+    // Permanent error, SC is not making any more transfer attempts
+    case 96: p = "Error,congestion"; break;
+    case 97: p = "Error,SME busy"; break;
+    case 98: p = "Error,no response sendr SME"; break;
+    case 99: p = "Error,service rejected"; break;
+    case 100: p = "Error,quality of service not available"; break;
+    case 101: p = "Error,error in SME"; break;
+    // 102...105: Reserved
+    // 106...111: Reserved
+    // 112...127: Values specific to each SC
+    // 128...255: Reserved
+
+    default:
+      if (status >= 48 && status <= 63)
+        p = "Temporary error, SC specific, unknown";
+      else if ((status >= 80 && status <= 95) || (status >= 112 && status <= 127))
+        p = "Permanent error, SC specific, unknown";
+  }
+
+  snprintf(dest, size_dest, "%s", p);
+}
+
 // Subroutine for messages type 0 (SMS-Deliver)
 // Input:
 // Src_Pointer points to the PDU string
@@ -1035,7 +1113,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
             }
 
             for (i = 0; sendr[i]; i++)
-              if (!isdigit(sendr[i]))
+              if (!isdigitc(sendr[i]))
               {
                 // Not a fatal error (?)
                 //pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding, "Invalid character(s) in sender address: \"%s\"", sendr);
@@ -1089,7 +1167,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
             {
               Src_Pointer += 2;
               sprintf(date,"%c%c-%c%c-%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(date[0]) || !isdigit(date[1]) || !isdigit(date[3]) || !isdigit(date[4]) || !isdigit(date[6]) || !isdigit(date[7]))
+              if (!isdigitc(date[0]) || !isdigitc(date[1]) || !isdigitc(date[3]) || !isdigitc(date[4]) || !isdigitc(date[6]) || !isdigitc(date[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in date of Service Centre Time Stamp: \"%s\"", date);
                 *date = 0;
@@ -1104,7 +1182,7 @@ int split_type_0(char *full_pdu, char* Src_Pointer, int* alphabet, char* sendr,
 
               Src_Pointer += 6;
               sprintf(time,"%c%c:%c%c:%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(time[0]) || !isdigit(time[1]) || !isdigit(time[3]) || !isdigit(time[4]) || !isdigit(time[6]) || !isdigit(time[7]))
+              if (!isdigitc(time[0]) || !isdigitc(time[1]) || !isdigitc(time[3]) || !isdigitc(time[4]) || !isdigitc(time[6]) || !isdigitc(time[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in time of Service Centre Time Stamp: \"%s\"", time);
                 *time = 0;
@@ -1180,7 +1258,6 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
   char tmpsender[100];
   int messageid;
   int i;
-  char *p;
   const char SR_MessageId[] = "Message_id:"; // Fixed title inside the status report body.
   const char SR_Status[] = "Status:"; // Fixed title inside the status report body.
 
@@ -1258,7 +1335,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
               }
 
               for (i = 0; sendr[i]; i++)
-                if (!isdigit(sendr[i]))
+                if (!isdigitc(sendr[i]))
                 {
                   // Not a fatal error (?)
                   //pdu_error(err_str, 0, Src_Pointer -full_pdu, Length +padding, "Invalid character(s) in recipient address: \"%s\"", sendr);
@@ -1278,7 +1355,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
             {
               // get SMSC timestamp
               sprintf(date,"%c%c-%c%c-%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(date[0]) || !isdigit(date[1]) || !isdigit(date[3]) || !isdigit(date[4]) || !isdigit(date[6]) || !isdigit(date[7]))
+              if (!isdigitc(date[0]) || !isdigitc(date[1]) || !isdigitc(date[3]) || !isdigitc(date[4]) || !isdigitc(date[6]) || !isdigitc(date[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in date of SMSC Timestamp: \"%s\"", date);
                 *date = 0;
@@ -1293,7 +1370,7 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
 
               Src_Pointer += 6;
               sprintf(time,"%c%c:%c%c:%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4]);
-              if (!isdigit(time[0]) || !isdigit(time[1]) || !isdigit(time[3]) || !isdigit(time[4]) || !isdigit(time[6]) || !isdigit(time[7]))
+              if (!isdigitc(time[0]) || !isdigitc(time[1]) || !isdigitc(time[3]) || !isdigitc(time[4]) || !isdigitc(time[6]) || !isdigitc(time[7]))
               {
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 6, "Invalid character(s) in time of SMSC Timestamp: \"%s\"", time);
                 *time = 0;
@@ -1327,8 +1404,8 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
             {
               // get Discharge timestamp
               sprintf(temp,"%c%c-%c%c-%c%c %c%c:%c%c:%c%c",Src_Pointer[1],Src_Pointer[0],Src_Pointer[3],Src_Pointer[2],Src_Pointer[5],Src_Pointer[4],Src_Pointer[7],Src_Pointer[6],Src_Pointer[9],Src_Pointer[8],Src_Pointer[11],Src_Pointer[10]);
-              if (!isdigit(temp[0]) || !isdigit(temp[1]) || !isdigit(temp[3]) || !isdigit(temp[4]) || !isdigit(temp[6]) || !isdigit(temp[7]) || 
-                  !isdigit(temp[9]) || !isdigit(temp[10]) || !isdigit(temp[12]) || !isdigit(temp[13]) || !isdigit(temp[15]) || !isdigit(temp[16]))
+              if (!isdigitc(temp[0]) || !isdigitc(temp[1]) || !isdigitc(temp[3]) || !isdigitc(temp[4]) || !isdigitc(temp[6]) || !isdigitc(temp[7]) || 
+                  !isdigitc(temp[9]) || !isdigitc(temp[10]) || !isdigitc(temp[12]) || !isdigitc(temp[13]) || !isdigitc(temp[15]) || !isdigitc(temp[16]))
                 pdu_error(err_str, 0, Src_Pointer -full_pdu, 12, "Invalid character(s) in Discharge Timestamp: \"%s\"", temp);
               else if (atoi(temp +3) > 12 || atoi(temp +6) > 31 || atoi(temp +9) > 24 || atoi(temp +12) > 59 || atoi(temp +16) > 59)
               {
@@ -1361,40 +1438,10 @@ int split_type_2(char *full_pdu, char* Src_Pointer,char* sendr, char* date,char*
                   pdu_error(err_str, 0, Src_Pointer -full_pdu, 2, "Invalid Status octet: \"%.2s\"", Src_Pointer);
                 else
                 {
-                  switch (status)
-                  {
-                    case 0: p = "Ok,short message received by the SME"; break;
-                    case 1: p = "Ok,short message forwarded by the SC to the SME but the SC is unable to confirm delivery"; break;
-                    case 2: p = "Ok,short message replaced by the SC"; break;
+                  char buffer[128];
 
-                    case 32: p = "Still trying,congestion"; break;
-                    case 33: p = "Still trying,SME busy"; break;
-                    case 34: p = "Still trying,no response sendr SME"; break;
-                    case 35: p = "Still trying,service rejected"; break;
-                    case 36: p = "Still trying,quality of service not available"; break;
-                    case 37: p = "Still trying,error in SME"; break;
-
-                    case 64: p = "Error,remote procedure error"; break;
-                    case 65: p = "Error,incompatible destination"; break;
-                    case 66: p = "Error,connection rejected by SME"; break;
-                    case 67: p = "Error,not obtainable"; break;
-                    case 68: p = "Error,quality of service not available"; break;
-                    case 69: p = "Error,no interworking available"; break;
-                    case 70: p = "Error,SM validity period expired"; break;
-                    case 71: p = "Error,SM deleted by originating SME"; break;
-                    case 72: p = "Error,SM deleted by SC administration"; break;
-                    case 73: p = "Error,SM does not exist"; break;
-
-                    case 96: p = "Error,congestion"; break;
-                    case 97: p = "Error,SME busy"; break;
-                    case 98: p = "Error,no response sendr SME"; break;
-                    case 99: p = "Error,service rejected"; break;
-                    case 100: p = "Error,quality of service not available"; break;
-                    case 101: p = "Error,error in SME"; break;
-
-                    default: p = "unknown";
-                  }
-                  sprintf(strchr(result, 0), "\n%s %i,%s", SR_Status, status, p);
+                  explain_status(buffer, sizeof(buffer), status);
+                  sprintf(strchr(result, 0), "\n%s %i,%s", SR_Status, status, buffer);
                 }
               }
             }
@@ -1549,7 +1596,7 @@ int splitpdu(char *pdu, char *mode, int *alphabet, char *sendr, char *date, char
                     if (smsc[Length -1] == 'F')
                       smsc[Length -1] = 0;
                     for (i = 0; smsc[i]; i++)
-                      if (!isdigit(smsc[i]))
+                      if (!isdigitc(smsc[i]))
                       {
                         // Not a fatal error (?)
                         //pdu_error(&err_str, 0, Pointer -pdu, Length, "Invalid character(s) in numeric SMSC address: \"%s\"", smsc);
